;;;			     mew-sort.el
;;;
;;;		 Copyright (C) 1994  Yoshinari NOMURA
;;;
;;;		   This emacs lisp library confirms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  Takashi P.KATOH <p-katoh@shiratori.riec.tohoku.ac.jp>
;;;          Kazuhiko Yamamoto <kazu@is.aist-nara.ac.jp>
;;; Created: Febraury  6, 1996
;;; Revised: 
;;;

(provide 'mew-sort)
(require 'mew)

;(define-key mew-summary-mode-map "Sd" 'mew-summary-sort-by-date)
;(define-key mew-summary-mode-map "Ss" 'mew-summary-sort-by-subject)
;(define-key mew-summary-mode-map "Sa" 'mew-summary-sort-by-author)
;(define-key mew-summary-mode-map "Sr" 'mew-summary-sort-by-recipient)
;(define-key mew-summary-mode-map "Sf" 'mew-summary-sort-by-field)


(defvar mew-summary-sort-last-field nil
  "Latest field used by mew-summary-sort-by-field.")

(defun mew-sort-body (method &rest options)
  (let ((folder (buffer-name)))
    (if (null (mew-summary-exclusive-p))
	()
      (if (mew-summary-mark-exist-p (list mew-mark-rmm mew-mark-refile))
	  (if (mew-y-or-n-p
	       "Marked messages exist. Process mark before scan? ")
	      (mew-summary-exec)))
      (if (null (mew-y-or-n-p (format "Sort %s by %s ? " folder method)))
	  ()
	(setq mew-summary-buffer-process t)
	(message "Sorting %s by %s ... " folder method)
	(apply 'call-process mew-prog-sortm nil nil nil folder options)
	(message "Sorting %s by %s ... done" folder method)
	(setq mew-summary-buffer-process nil)
	(let ((buffer-read-only nil)) (erase-buffer))  ;; for update
	(mew-summary-scan-body folder (mew-input-range folder))
	)
      )
    ))

(defun mew-summary-sort-by-date ()
  (interactive)
  (mew-sort-body "date" "-noverbose"))

(defun mew-summary-sort-by-subject ()
  (interactive)
  (mew-sort-body "subject" "-textfield" "subject" "-limit" "0"))

(defun mew-summary-sort-by-author ()
  (interactive)
  (mew-sort-body "author" "-textfield" "from" "-limit" "0"))

(defun mew-summary-sort-by-recipient ()
  (interactive)
  (mew-sort-body "recipient" "-textfield" "to" "-limit" "0"))

(defun mew-summary-sort-by-field (&optional field)
  (interactive)
  (let ((tmp (read-input
	      (concat "Field"
		       (if mew-summary-sort-last-field
			   (concat " (default " mew-summary-sort-last-field "): ")
			 ": ")))))
    (setq tmp (if (string-equal tmp "") nil tmp))
    (or field (setq field (or tmp
			      mew-summary-sort-last-field)))
    (setq mew-summary-sort-last-field field)
    (mew-sort-body field "-textfield" field "-limit" "0")))

