;;;			    mew-charset.el
;;;
;;;		Copyright (C) 1996  Kazuhiko Yamamoto
;;;
;;;		   This emacs lisp library conforms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  Kazuhiko Yamamoto <kazu@is.aist-nara.ac.jp>
;;; Created: October   2, 1996
;;; Revised: November 15, 1996
;;;

(defconst mew-charset-version "mew-charset.el version 0.03")

(require 'mew)

(if (null mew-mule-p)
    ()
  (defvar mew-mule-character-set
    (list 
     (if (boundp '*iso-8859-1*)	;; Latin-1
	 (list lc-ltn1	*iso-8859-1*		"iso-8859-1")
       (if (boundp '*ctext*) ;; for old mule
	   (list lc-ltn1	*ctext*		"iso-8859-1")))
     (if (boundp '*iso-8859-2*)	;; Latin-2
	 (list lc-ltn2	*iso-8859-2*		"iso-8859-2"))
     (if (boundp '*iso-8859-3*)	;; Latin-3
	 (list lc-ltn3	*iso-8859-3*		"iso-8859-3"))
     (if (boundp '*iso-8859-4*)	;; Latin-4
	 (list lc-ltn4	*iso-8859-4*		"iso-8859-4"))
     (if (boundp '*iso-8859-5*)	;; Cyrillic
	 (list lc-crl		*iso-8859-5*	"iso-8859-5"))
     (if (boundp '*iso-8859-6*)	;; Arabic
	 (list lc-arb		*iso-8859-6*	"iso-8859-6"))
     (if (boundp '*iso-8859-7*)	;; Greek
	 (list lc-grk		*iso-8859-7*	"iso-8859-7"))
     (if (boundp '*iso-8859-8*)	;; Hebrew
	 (list lc-hbw		*iso-8859-8*	"iso-8859-8"))
     (if (boundp '*iso-8859-9*)	;; Latin-5
	 (list lc-ltn5	*iso-8859-9*		"iso-8859-9"))
     (if (boundp '*iso-2022-jp*);; Japanese
	 (list lc-jp		*iso-2022-jp*	"iso-2022-jp"))
     (if (boundp '*iso-2022-kr*);; Korean
	 (list lc-kr		*iso-2022-kr*	"iso-2022-kr"))
     (if (boundp '*iso-2022-ss2-7*);; Chinese xxx
	 (list lc-cn		*iso-2022-ss2-7* "iso-2022-cn"))
     (if (boundp '*iso-2022-ss2-7*)
	 (list t		*iso-2022-ss2-7* "iso-2022-jp-2"))
     ;; for charset to symbol
     (if (boundp '*euc-japan*)
	 (list nil           *euc-japan*	  "euc-japan")) 
     (if (boundp '*euc-korea*)
	 (list nil           *euc-korea*          "euc-korea"))
     (if (boundp '*sjis*)
	 (list nil           *sjis*               "sjis")) 
     ))
  (defvar mew-mule-mime-charset 
    (mapcar (function (lambda (x) (nth 2 x))) mew-mule-character-set))
  )

(defmacro mew-mule-lc-attr (lc alist)
  (` (mew-assoc (, lc) (, alist) 0 nil)))

(defmacro mew-mule-lc-attr-by-charset (lc alist)
  (` (mew-assoc (, lc) (, alist) 2 t)))

(defmacro mew-mule-lc-content (attr)
  (` (car (cdr (cdr (, attr))))))

(defmacro mew-mule-lc-symbol (attr)
  (` (car (cdr (, attr)))))

(defun mew-charset-to-symbol (charset)
  (mew-mule-lc-symbol 
   (mew-mule-lc-attr-by-charset charset mew-mule-character-set)))

(defmacro mew-mule-content-attr (content alist)
  (` (mew-assoc2 (, content) (, alist) 2 t)))

(defmacro mew-mule-content-coding (attr)
  (` (car (cdr (, attr)))))

(cond
 (mew-mule-p
  (defvar mew-mule-charset-mime-trans   *iso-2022-ss2-7*) 
  (defvar mew-mule-charset-rfc822-trans *iso-2022-ss2-7*) 
  (defvar mew-mule-charset-draft        *iso-2022-jp*)
  (defvar mew-mule-charset-scan         *iso-2022-jp*) ;; *iso-2022-ss2-7*
  (defvar mew-mule-charset-spool        *noconv*)
  (defvar mew-mule-charset-infile       *autoconv*)
  (defvar mew-mule-charset-outfile      *iso-2022-jp*)
  (defvar mew-mule-charset-virtual      *iso-2022-jp*unix) ;; ^M as it is
  )
 )

(cond 
 (mew-mule-p
  (define-program-coding-system nil mew-prog-inc  mew-mule-charset-scan)
  (define-program-coding-system nil mew-prog-scan mew-mule-charset-scan)
  (define-program-coding-system nil mew-prog-send 
    (cons *noconv* mew-mule-charset-mime-trans))
  (define-program-coding-system nil mew-prog-pick mew-mule-charset-scan)
  (define-program-coding-system nil mew-prog-mime-encode *noconv*)
  (define-program-coding-system nil mew-prog-mime-decode *noconv*)
  (define-program-coding-system nil mew-prog-vscan mew-mule-charset-virtual)
  )
 )

(defun mew-charset-8bit-p (charset)
  (let ((case-fold-search t))
    (cond 
     ((string-match "iso-8859" charset) t)
     ((string-match "iso-2022" charset) nil)
     ((string-match "us-ascii" charset) nil)
     (t t) ;; just in case
     )
    ))

(defun mew-charset-iso-2022-p (charset)
  (let ((case-fold-search t))
    (cond 
     ((string-match "iso-2022" charset) t)
     (t nil)
     )
    ))

(defun mew-charset-guess-region (beg end)
  (interactive "r")
  "Guess minimum character set name."
  (cond 
   (mew-mule-p
    (let* ((lc (find-charset-region beg end))
	   (len (length lc)))
      (cond
       ((equal len 0) "us-ascii")
       ((equal len 1) (mew-mule-lc-content 
		       (mew-mule-lc-attr (car lc) mew-mule-character-set)))
       (t "iso-2022-jp-2")
       )))
   (mew-emacs19-p
    (save-excursion
      (goto-char beg)
      (if (re-search-forward "[\200-\377]" end t)
	  "iso-8859-1"
	"us-ascii")))
   (t "us-ascii") ;; Emacs
   ))

(provide 'mew-charset)
