;;;	     mew.el :: Message interface to Emacs Window
;;;
;;;	  Copyright (C) 1994, 1995, 1996  Kazuhiko Yamamoto
;;;
;;;		   This emacs lisp library conforms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  Kazuhiko Yamamoto <kazu@is.aist-nara.ac.jp>
;;; Created: March    22, 1994
;;; Revised: November 25, 1996
;;;

(defconst mew-version "Mew version 1.54")
(provide 'mew)


;;;  Minimum setup.
;;;
;;; (autoload 'mew "mew" nil t)
;;; (autoload 'mew-send "mew" nil t)

(defvar mew-debug nil)
;(setq mew-debug nil)
;(setq mew-debug t)

(defvar mew-emacs18-p (string-match "^18" emacs-version))
(defvar mew-emacs19-p (string< "19" emacs-version))
(defvar mew-mule-p    (boundp 'MULE))
(defvar mew-xemacs-p  (string-match "XEmacs\\|Lucid" emacs-version))
(defvar mew-os2-p     nil)

(defvar mew-profile (or (getenv "MH") (expand-file-name "~/.mh_profile")))

;;;
;;; user customizable variables
;;;

(defvar mew-analysis t
  "*MIME syntax analysis and cache when non-nil.
   Skip syntax analysis when nil.")

(defvar mew-cache-size 10
  "*Number of buffer for message cache.")

(defvar mew-summary-cache-file ".mew-cache"
  "*Cache file for summary mode contents.")

(defvar mew-summary-cache-use t
  "*If non-nil, insert mew-summary-cache-file into summary mode buffer 
when the summary mode is visited and it does not exist in buffer. 
mew-summary-cache-file is automatically saved on inc and scan.")

(defvar mew-window-use-full nil
  "*Dynamic window configuration when nil. This variable will be obsolated
because mew will support mew-window-percent(e.g 50, 100, etc.).")

(defvar mew-summary-show-direction 'down
  "*Direction for SPC at end of message. 
'up 'down 'next(current direction) 'stop")

(defvar mew-mime-compose-folder-delete 'ask
 "*Action to delete MIME draft directory after sending a multipart letter.
Proper value is 'ask', 'delete', and 'retain'.")


(defvar mew-mail-domain-list nil
  "*Your e-mail address domain list like 
(\"is.aist-nara.ac.jp\" \"csce.kyushu-u.ac.jp\").
They are used for mail-domain completion in draft mode(C-cTAB)")

(defvar mew-mail-domain (or (car mew-mail-domain-list) "")
  "*Your e-mail address domain.")

(defvar mew-mail-address-list nil
  "*The addresses included in this list never appear on the Cc:
field on a draft buffer.")


(defvar mew-redist-needs-full-contents nil)
  ; if you are using MH/post that speaks SMTP, leave this as "nil".
  ; if you are using MH/post that spawns sendmail and passes message
  ; with PIPE, change this as t.

(defvar mew-eof-string nil
  "*A value inserted on the end of message buffer if non-nil.")

;; Draft fields

(defvar mew-reply-to nil
  "*A value inserted into Reply-To: field in draft mode if non-nil.")
(defvar mew-fcc nil
  "*A value inserted into Fcc: field in draft mode if non-nil.")
(defvar mew-cc nil
  "*A value inserted into Cc: field in draft mode if non-nil.")
(defvar mew-dcc nil
  "*A value inserted into Dcc: field in draft mode if non-nil.")
(defvar mew-from nil
  "*A value inserted into From: field in draft mode if non-nil.")
(defvar mew-x-mailer
  (concat mew-version " on Emacs " emacs-version
	  (and mew-mule-p
	       (string-match "^[0-9.]*" mule-version)
	       (concat ", Mule " (substring mule-version
					    (match-beginning 0) 
					    (match-end 0)))))
  "*A value inserted into From: field in draft mode if non-nil.")

(defvar mew-header-alist nil
  "*alist of key and value pair for header field to be inserted on draft.
'((\"X-fingerprint:\" . \"6B 63 38 88 67 5E 96 8E  CE A4 62 73 3F 11 64 94\")
  (\"X-URL:\" . \"http://uranus.aist-nara.ac.jp/~kazu/\"))"
  )

(defvar mew-auto-add-content-type nil
  "*Automatically adds MIME Content-Type: header if non-nil, asks otherwise.")
(defvar mew-ask-cc t
  "*Prompt user for CC: field if non-nil")
(defvar mew-ask-subject nil
  "*Prompt user for Subject: field when send if non-nil")
(defvar mew-ask-range t
  "*If non-nil, ask for a range for scanning. If false, always
use default.")



(defvar mew-emacs-y-or-n-p-use nil
  "*SPC is y and RET is n if t. SPC is n and RET is y, otherwise.")

;; Overlays for mouse highlight and cursor underline

(defvar mew-highlight-lines-use nil
  "*Use highlight line in summary mode.")
(defvar mew-underline-lines-use mew-emacs19-p
  "*Use underline line in summary mode.")
(defvar mew-bold-url-size 3000
  "*Use bold font at url maximize size.
If nil, all contents are parsed.")

(cond 
 (mew-xemacs-p
  (or (find-face 'underline)
      (progn (make-face 'underline)
	     (set-face-underline-p 'underline t)))
  (fset 'mew-overlayp (symbol-function 'extentp))
  (fset 'mew-make-overlay (symbol-function 'make-extent))
  (defun mew-move-overlay (overlay beg end &optional buffer)
    (set-extent-endpoints overlay beg end))
  (defun mew-overlay-put (overlay prop value)
    (set-extent-property overlay prop value))
  (require 'highlight-headers)
  )
 (mew-emacs19-p
  ;; to avoid competition with mh-e.el
  (if (rassq 'mh-letter-mode auto-mode-alist)
      (setq auto-mode-alist
	    (delete (rassq 'mh-letter-mode auto-mode-alist)
		    auto-mode-alist)))
;;  (setq auto-mode-alist
;;	(cons '("/drafts/[0-9]+\\'" . mew-draft-mode) auto-mode-alist))
  (add-hook 'kill-emacs-hook (function mew-clean-up))
  (fset 'mew-overlayp (symbol-function 'overlayp))
  (fset 'mew-make-overlay (symbol-function 'make-overlay))
  (fset 'mew-move-overlay (symbol-function 'move-overlay))
  (fset 'mew-overlay-put (symbol-function 'overlay-put))
  )
 )

;; Only one signature file is supported in Mew. 
;; Please use c-sig.el for multiple signatures.
(defvar mew-signature-file "~/.signature")

(defvar mew-signature-insert-last nil 
  "If non-nil, the signature file is inserted in the last of body. 
Otherwise, it is inserted at the currect point. If you created multipart
and mew-signature-as-lastpart is non-nil, this variable is ignored.")

(defvar mew-signature-as-lastpart nil
  "If non-nil, the signature file is appended as the final part 
when you create multipart.")

(defvar mew-signature-description "My signature"
  "This variable is used as a description if the signature is appended
as the final part.")

(defvar mew-file-append-p nil
  "If non-nil, a message or a part is appended to the existing file
on mew-summary-save(\"y\"). Otherwise overwrited.")

(defvar mew-save-application/octet-stream-p t)

;;;
;;; Citation
;;;

(defvar mew-cite-prefix "> "
  "*Prefix of citation.")
(defvar mew-cite-hook nil
  "*If you want to use super-cite, (setq mew-cite-hook 'sc-cite-original).")

(defvar mew-cite-prefix-function nil)
;(defvar mew-cite-prefix-function 'mew-cite-prefix-username)

(defvar mew-cite-prefix-confirmp nil)
(defun mew-cite-prefix-username ()
  (format "%s> " (mew-header-delete-at
		  (mew-header-extract-addr
		   (mew-header-get-value "From:"))))
  )

(defvar mew-cite-fields '("From:" "Subject:" "Date:")
  "*The fields that you want to extract as citation label. 
If you change this valuable, you must change mew-cite-format.
The value of the first field becomes the first argment for mew-cite-format.
(e.g. The default first argment is a value of From: field.)
The value of the second field becomes the second argment for mew-cite-format.
....")

(defvar mew-cite-format "From: %s\nSubject: %s\nDate: %s\n\n")

;;;
;;; Hooks
;;;

(defvar mew-init-hook nil
  "*Hook called at initialize time.")
(defvar mew-summary-mode-hook nil
  "*Hook called in summary mode.")
(defvar mew-draft-mode-hook nil
  "*Hook called in draft mode.")
(defvar mew-message-mode-hook nil
  "*Hook called in message mode.")
(defvar mew-message-hook nil
  "*Hook called whenever message displayed.")
(defvar mew-send-hook nil
  "*Hook called before sending a message")
(defvar mew-suspend-hook nil
  "*Hook called on suspend.")
(defvar mew-quit-hook nil
  "*Hook called on quit.")
(defvar mew-summary-inc-sentinel-hook nil
  "*Hook called when inc finished.")
(defvar mew-summary-scan-sentinel-hook nil
  "*Hook called when scan finished.")
(defvar mew-summary-exec-hook nil
  "*Hook called when mew-summary-exec finished.")

(defvar mew-summary-recenter-p t)
(defvar mew-break-pages t)
(defvar mew-page-delimiter "^\^L")

;;
;; BBDB not fully supported
;;

(defvar mew-use-bbdb nil
  "*Use Insidious Big Brother Database if t")


(defvar mew-inbox "+inbox")
(defvar mew-trash-folder nil
  "*If non-nil, messages to be deleted are refiled to the defined folder")
(defvar mew-folders-file ".folders")
(defvar mew-folders-ignore '("from"))
(defvar mew-folders-default-folder "from")

(defvar mew-print-function (function lpr-buffer))

(defvar mew-default-boundary "--%s(%s)--")

(defvar mew-x-mew: "X-Mew:")

(cond 
 ((string= mew-mail-domain "")
  (defvar mew-mail-address (user-login-name)))
 (t
  (defvar mew-mail-address (concat (user-login-name) "@" mew-mail-domain)))
 )

;;
;; It's nice to tell current message if lines of summary window is odd.
;;
(defvar mew-window-configuration 
  (if mew-emacs19-p
      '((summary (1  0))
	(message (8 31)))
    '((summary (1  0))
      (message (4 19)))
    )
  "*Ratio of windows")

(defvar mew-summary-message-regex "^ *\\([0-9]+\\)")
(defvar mew-summary-part-regex "^\t+\\([0-9.]+\\)")
(defvar mew-address-separator ":, \t\n")
(defvar mew-summary-edit-again-regex
  "----- Original message follows -----\\|----- Unsent message follows -----")

;; X face

(defvar mew-x-face-filter '("uncompface" "ikon2xbm"))
(defvar mew-x-face-prog "xv")
(defvar mew-x-face-args nil)
(defvar mew-x-face-file "~/.xface"
  "*If non-nil and the file exists, X-Face: fields is inserted.")

;;
;; current solution to decide To: when reply
;;

(defvar mew-noreplyto-to-list '("From:")
  "To: field list picked at mew-summary-reply if Reply-To: doesn't exist.")

(defvar mew-noreplyto-cc-list '("To:" "Cc:" "Apparently-To:")
  "Cc: field list picked at mew-summary-reply if Reply-To: doesn't exist.")

(defvar mew-replyto-to-list '("Reply-To:" "From:")
  "To: field list picked at mew-summary-reply if Reply-To: exists.")

(defvar mew-replyto-cc-list '("To:" "Cc:" "Apparently-To:")
  "Cc: field list picked at mew-summary-reply if Reply-To: exists.")

;;
;;

;; If you change the order or add new fields, change mew-part-*.

(defvar mew-prog-inc         "inc")
(defvar mew-prog-scan        "scan")
(defvar mew-prog-vscan       "vscan")
(defvar mew-prog-ali         "ali")
(defvar mew-prog-folders     "folders")
(defvar mew-prog-folder      "folder")
(defvar mew-prog-pack        "folder")
(defvar mew-prog-rmm         "rmm")
(defvar mew-prog-refile      "refile")
(defvar mew-prog-send        "send")
(defvar mew-prog-sortm       "sortm")
(defvar mew-prog-pick        "pick")
(defvar mew-prog-msgchk      "msgchk")
(defvar mew-prog-mime-encode "mewencode")
(defvar mew-prog-mime-decode "mewdecode")

(defvar mew-prog-inc-arg-list nil
  "Should be argment LIST!")
;;
;; Oooops, mew-summary-scan-body applies range!
;;(defvar mew-prog-scan-arg-list nil
;;  "Should be argment LIST!")

(defvar mew-prog-unshar   "unshar")
(defvar mew-prog-uumerge  "uumerge")
(defvar mew-prog-tar      "tar")
(defvar mew-prog-compress "compress")
(defvar mew-prog-gzip     "gzip")

(defvar mew-mark-rmm    ?D)
(defvar mew-mark-mark   ?@)
(defvar mew-mark-hop    ?*)
(defvar mew-mark-refile ?o) ;; do you like "^"?
(defvar mew-mark-tmp    ?%) ;; temporary use only.

(defvar mew-temp-file-initial (expand-file-name (user-login-name) "/tmp")
  "Hint to make a secure directory on the local file system.
On setup phase Mew make a secure directory from this variable
and set mew-temp-file a file name prefex contained the directory name.
The directory must be unreadable from others, otherwise it might become 
a big security hole. Mew never use 'call-process-region' rather does use
'call-process' creating a temporary file with mew-temp-file by itself.
If 'call-process-region' is used, Emacs/Mule creates a temporary file
(probably in /tmp). So bad guys can wiretap the temporary file.")

(setq mew-temp-dir nil)  ;; the default is "/tmp/user_name_uniq"
(setq mew-temp-file nil) ;; the default is "/tmp/user_name_uniq/mew"
(defvar mew-prog-chmod "chmod")
(defvar mew-prog-chmod-arg "700")

(defvar mew-mark-switch
  (list 
   (cons mew-mark-rmm    'mew-delete-line)
   (cons mew-mark-refile 'mew-delete-line)
   (cons mew-mark-mark   'mew-summary-mark-letitbe)
   (cons mew-mark-hop    'mew-summary-mark-letitbe)
   )
  )

(defvar mew-undo-switch
  (list 
   (cons mew-mark-rmm    'mew-summary-unmark)
   (cons mew-mark-refile 'mew-summary-refile-undo)
   (cons mew-mark-mark   'mew-summary-unmark)
   (cons mew-mark-hop    'mew-summary-unmark)
   )
  )

;(defvar mew-draft-mime-folder (format "+%s/mime" mew-draft-folder))
(defvar mew-draft-mime-folder "+drafts/mime")

(defvar mew-minibuffer-map      nil)
(defvar mew-folder-map          nil)

(defvar mew-header-decode-regex nil)

;;;
;;; filename "^$" never matches anything
;;;

(defvar 
 mew-mime-content-type
 '(
   ("audio/basic"
    "\\.au$"	"base64"	("sh" ("-c" "cat - > /dev/audio") nil))
;   ("audio/basic"
;    "\\.au$"	"base64"	("aplay" () t))
;   ("audio/basic"
;    "\\.au$"	"base64"	("play_au" () t)) ;; BSD/OS
   ("image/gif"
    "\\.gif$" 	"base64"	("xv" ("-geometry" "+0+0") t))
   ("image/jpeg"
    "\\.jpe?g$"	"base64"	("xv" ("-geometry" "+0+0") t))
   ("image/x-xwd"
    "\\.xwd$"	"base64"	("xv" ("-geometry" "+0+0") t))
   ("image/x-xbm"
    "\\.xbm$"	"base64"	("xv" ("-geometry" "+0+0") t))
   ("image/x-bmp"
    "\\.bmp$"	"base64"	("xv" ("-geometry" "+0+0") t))
   ("image/.*"
    "^$"        "base64"	("xv" ("-geometry" "+0+0") t))
   ("video/mpeg"
    "\\.mpe?g$"	"base64"	("mpeg_play" () t))
   ("application/postscript"
    "\\.ps$"	"quoted-printable"	("ghostview" ("-geom" "+0+0") t))
   ("application/pgp-keys"
    "\\.pka$"   nil	        (mew-mime-pgp-keys () nil))
   ("application/octet-stream"
    "\\.tar\\.?g?z?$" "base64"	(mew-mime-application/octet-stream () nil))
   ("application/octet-stream"
    "\\.gz$" "base64"           (mew-mime-application/octet-stream () nil))
   ("application/octet-stream"
    "\\.Z$" "base64"	        (mew-mime-application/octet-stream () nil))
   ("application/x-graph"
    "\\.xg$"	"quoted-printable"	("xgraph" () t))
   ("message/external-body" 
    "\\.ext$"	nil		(mew-mime-external-body () nil))
   ("message/rfc822"
    "^[0-9]+$"	nil		(mew-mime-message/rfc822 () nil))
   ("message/delivery-status"
    "^$"	nil		(mew-mime-text/plain () nil))
;   ("text/html"
;    "\\.html$"	nil		(mew-mime-text/html () nil))
   ("text/richtext"
    "\\.rtf$"	nil		(mew-mime-text/plain () nil))
   ("text/plain"
	".*"	nil		(mew-mime-text/plain () nil))
   ("text/.*"
	"^$"    nil             (mew-mime-text/plain () nil))
   (".*" 
        "^$"    nil             (mew-mime-application/octet-stream () nil))
   )
 "(content-type filename encoding (program (options ...) async))"
 )

(defvar mew-mime-content-type-text-list
  '("Text/Plain" "Text/Html"
    "Application/Postscript")
  "Content-Type: list for text which are treated as text when encoded"
  )

(defvar mew-mime-content-type-multipart-list
  '("Multipart/Mixed" "Multipart/Alternative"
    ;;"Multipart/Digest" "Multipart/Parallel"
    )
  "Candidate of 'Content-Type: Multipart' when CT: is changed in draft buffer."
  )

(defvar mew-mime-content-type-binary-list
  '("Application/Octet-Stream"
    "Image/Jpeg" "Image/Gif" "Image/x-xwd" "Image/x-xbm" "Image/x-bmp"
    "Audio/Basic" "Video/Mpeg"
    )
  "Content-Type: list to save as binary file in mew-summary-save on Mule"
  )

(defvar mew-mime-content-type-others-list
  '("Message/Rfc822" "Message/External-body"
    "Application/X-Graph")
  )

(defvar mew-mime-content-type-list
  (append mew-mime-content-type-text-list
	  mew-mime-content-type-multipart-list
	  mew-mime-content-type-binary-list
	  mew-mime-content-type-others-list)
  "Candidate of Content-Type: when CT: is changed in draft buffer."
  )

;;
(defmacro mew-file-attr (filename alist)
  (` (mew-assoc2 (, filename) (, alist) 1 t)))

(defmacro mew-type-attr (filename alist)
  (` (mew-assoc2 (, filename) (, alist) 0 nil)))

;; for mapcar
(defun mew-file-content (attr)
  (car attr))

(defmacro mew-file-encoding (attr)
  (` (car (cdr (cdr (, attr))))))

;;
(defmacro mew-content-attr (content alist)
  (` (mew-assoc2 (, content) (, alist) 0 t)))

(defmacro mew-content-program (attr)
  (` (car (car (cdr (cdr (cdr (, attr))))))))

(defmacro mew-content-options (attr)
  (` (car (cdr (car (cdr (cdr (cdr (, attr)))))))))


(defmacro mew-content-async (attr)
  (` (car (cdr (cdr (car (cdr (cdr (cdr (, attr))))))))))

;;;
;;;
;;;

;;;
;;; constant
;;;

(defconst mew-buffer-message     "*Mew message*")
(defconst mew-buffer-mime       " *mew mime*")
(defconst mew-buffer-hello      " *mew hello*")
(defconst mew-buffer-cache      " *mew cache*")
(defconst mew-buffer-watch       "*Mew watch*")
(defconst mew-buffer-tmp        " *mew tmp*")
(defconst mew-buffer-completions "*Mew completions*")
(defconst mew-buffer-whom        "*Mew whom*")


;; set here due to mew-buffer-tmp
(defvar mew-prog-audio (list "cat" "/dev/audio" mew-buffer-tmp nil))
;(defvar mew-prog-audio '("arecord" nil nil nil "-silentlevel" "-100"))

(defun mew-buffer-message ()
  (if (and window-system mew-emacs19-p)
      (concat
       mew-buffer-message
       (int-to-string
	(if mew-xemacs-p
	    (mew-member-match 
	     (cdr (assq
		   'window-id
		   (screen-parameters (window-screen (selected-window)))))
	     (sort
	      (mapcar
	       (function 
		(lambda (frame)
		  (cdr (assq
			'window-id
			(screen-parameters frame)))))
	       (screen-list))
	      (function string<)))
	  (mew-member-match 
	   (cdr (assq
		 'window-id
		 (frame-parameters (window-frame (selected-window)))))
	   (sort
	    (mapcar
	     (function 
	      (lambda (frame)
		(cdr (assq
		      'window-id
		      (frame-parameters frame)))))
	     (frame-list))
	    (function string<))))))
    mew-buffer-message
    ))


(defconst mew-hello-message 
"

Welcome to Mew world.

Mew -- Message interface to Emacs Window

%s

Copyright (C) 1994, 1995, 1996  Kazu Yamamoto

Please send comments to kazu@is.aist-nara.ac.jp.

/\\\\ - \\\\/"
)

(defconst mew-error-message 
"

    Sorry, your configuration is *NOT* appropriate for Mew. 
    Please check the following items.

        (1) Does \"~/.mh_profile\" exist?
        (2) Does \"~/.mh_profile\" contain \"Path: Mail\" field?
        (3) Does \"~/.mh_profile\" contain \"Draft-Folder: drafts\" field?
        (4) Does \"~/Mail/drafts\" exist?
        (5) Is the path for MH included in exec-path?
              Put (setq exec-path (cons \"/usr/local/bin/mh\" exec-path))
              in .emacs.
	(6) Is the version of mewencode 0.08?
")

;;;
;;; initialize global variables
;;;

(setplist 'mew-current-state 
	  (list
	   'message    nil 
	   'cache      nil
	   'part       nil
	   'window     nil
	   ))

(defun mew-current-get (key)
  (get 'mew-current-state key))

(defun mew-current-set (key value)
  (put 'mew-current-state key value))

(defvar mew-path nil) ;; should be nil

(setq mew-draft-folder nil)
(setq mew-alias-alist  ())
(setq mew-folder-list  ())
(setq mew-folder-alist ())
(setq mew-refile-alist ())
(setq mew-folders (list mew-inbox))

(setq mew-message-citation nil)
(setq mew-message-citation-buffer nil) ;; should be nil
(defvar mew-last-shell-command "")

(setq mew-process-file-alist nil)

(defvar mail-header-separator "----")


;;
;; buffer local
;;

(defvar mew-summary-buffer-process nil)
(defvar mew-summary-buffer-string nil)
(defvar mew-summary-buffer-direction nil)
(defvar mew-summary-buffer-folder-cache-time nil)
(defvar mew-watch-buffer-process nil)
(defvar mew-draft-buffer-header nil)
(defvar mew-draft-buffer-attach nil)
(defvar mew-decode-syntax ()
  "Variable to save MIME syntax list")
(defvar mew-summary-buffer-refile nil)
(defvar mew-cache-attribute nil)
(defvar mew-summary-buffer-disp-msg t)
(defvar mew-summary-buffer-left-msgs "-")
(defvar mew-encode-syntax nil)
(defvar mew-underline-overlay nil)

(mapcar
 (function make-variable-buffer-local)
 (list 'mew-summary-buffer-process
       'mew-summary-buffer-string
       'mew-summary-buffer-direction
       'mew-summary-buffer-folder-cache-time
       'mew-encode-syntax
       'mew-decode-syntax
       'mew-summary-buffer-refile
       'mew-cache-attribute
       'mew-watch-buffer-process
       'mew-draft-buffer-header
       'mew-draft-buffer-attach
       'mew-summary-buffer-disp-msg
       'mew-summary-buffer-left-msgs
       'mew-underline-overlay
       ))

;;;
;;; environment setting
;;;

(cond 
 (mew-emacs18-p
  (defun mew-menu-define (symbol maps doc menu))
  (fset 'mew-buffer-substring 'buffer-substring))
 (mew-xemacs-p
  (defmacro mew-menu-define (symbol maps doc menu)
    (` (progn
	 (defvar (, symbol) nil (, doc))
	 (mew-menu-do-define
	  (quote (, symbol)) (, maps) (, doc) (, menu)))))
  (defun mew-menu-do-define (symbol maps doc menu)
    (set symbol menu)
    (fset symbol (list 'lambda '(e)
		       doc
		       '(interactive "@e")
		       '(run-hooks 'activate-menubar-hook)
		       '(setq zmacs-region-stays 't)
		       (list 'popup-menu symbol)))
    (define-key maps 'button3 symbol))
  (fset 'mew-buffer-substring 'buffer-substring))
 (t ;; mew-emacs19-p not xemacs
  (require 'easymenu)
  (fset 'mew-menu-define 'easy-menu-define)
  (if (fboundp 'buffer-substring-no-properties)
      (fset 'mew-buffer-substring 'buffer-substring-no-properties)
    (defun mew-buffer-substring (beg end)
      "Return the text from BEG to END, without text properties, as a string."
      (let ((string (buffer-substring beg end)))
	(set-text-properties 0 (length string) nil string)
	string)))
  )
 )

(if mew-minibuffer-map
    ()
  (setq mew-minibuffer-map (make-sparse-keymap))
  (define-key mew-minibuffer-map "\t"     'mew-minibuffer-alias)
  (define-key mew-minibuffer-map "\C-c\t" 'mew-minibuffer-domain)
  (define-key mew-minibuffer-map "\r"     'exit-minibuffer)
  (define-key mew-minibuffer-map "\n"     'exit-minibuffer)
  (define-key mew-minibuffer-map "\C-g"   'abort-recursive-edit)
  (if mew-use-bbdb
      (define-key mew-minibuffer-map "\e\t"     'bbdb-complete-name))
  )

(if mew-folder-map
    ()
  (setq mew-folder-map (make-sparse-keymap))
  (define-key mew-folder-map " "    'mew-minibuffer-folders-word)
  (define-key mew-folder-map "\t"   'mew-minibuffer-folders)
  (define-key mew-folder-map "\r"   'exit-minibuffer)
  (define-key mew-folder-map "\n"   'exit-minibuffer)
  (define-key mew-folder-map "\C-g" 'abort-recursive-edit)
  )

;;;
;;; Basic functions
;;;

(defun mew-version-show ()
  (interactive)
  (message "%s" mew-version))

(defun mew-replace-character (string from to)
  (let ((len (length string))
	(cnt 0))
    (while (< cnt len)
      (if (equal (aref string cnt) from)
	  (aset string cnt to))
      (setq cnt (1+ cnt)))
    string
    ))

(defun mew-y-or-n-p (string)
  (interactive)
  (if mew-emacs-y-or-n-p-use
      (y-or-n-p string)
    (let ((char nil) (ociea cursor-in-echo-area))
      (unwind-protect
	  (progn
	    (setq cursor-in-echo-area t)
	    (catch 'break
	      (while t
		(message (concat string "(y=RET or n=SPC) "))
		(setq char (read-char))
		(cond
		 ((equal char ?y)
		  (setq char t)
		  (throw 'break nil))
		 ((equal char ?\r)
		  (setq char t)
		  (throw 'break nil))
		 ((equal char ?n)
		  (setq char nil)
		  (throw 'break nil))
		 ((equal char 32)
		  (setq char nil)
		  (throw 'break nil))
		 (t (ding))
		 )
		))
	    )
	(setq cursor-in-echo-area ociea)
	)
      char ;; return value
      ))
  )

(defun mew-member-match (str list)
  (let ((n 0))
    (catch 'member
      (while list
	(if (equal (downcase (car list)) (downcase str))
	    (throw 'member n))
	(setq list (cdr list))
	(setq n (1+ n))
	))
    ))

(defun mew-delq (key list)
  (let* ((pointer (cons nil list))
	 (top pointer))
    (while (cdr pointer)
      (if (equal key (car (cdr pointer))) 
	  (progn
	    (setcdr pointer (cdr (cdr pointer)))
	    (setq pointer (cons nil nil)))
	(setq pointer (cdr pointer))))
    (cdr top)))

(defun mew-folder-to-dir (folder)
  (if (equal (aref folder 0) ?+)
      (substring folder 1 nil)
    folder)
  )

(defun mew-dir-to-folder (dir)
  (if (equal (aref dir 0) ?+)
      dir
    (concat "+" dir))
  )

(defmacro mew-draft-to-mime (draft)
  (` (concat mew-draft-mime-folder "/" (file-name-nondirectory (, draft)))))

(defun mew-make-folder (folder)
  (mew-make-directory (mew-folder-to-dir folder)))

(defun mew-make-directory (path)
  (let ((parent (directory-file-name (file-name-directory path))))
    (if (null (file-directory-p parent))
	(mew-make-directory parent))
    (if (and (file-exists-p path) (not (file-directory-p path)))
	(delete-file path))
    ;; Emacs18 doesn't have make-directory
    (call-process "mkdir" nil nil nil path)
    ))

;; I don't like such a procedural programming, but for max depth safety.
(defun mew-rassoq (a alist)
  (catch 'conscell
    (while alist
      (if (equal (cdr (car alist)) a)
	  (throw 'conscell (car alist)))
      (setq alist (cdr alist)))))

(defun mew-match (pos &optional string)
  (cond 
   ((stringp string)
    (substring string (match-beginning pos) (match-end pos)))
   (t 
    (mew-buffer-substring (match-beginning pos) (match-end pos)))
   ))

(defun mew-assoc (key alist nth match)
  (let ((case-fold-search t))
    (if match
	(mew-assoc-match key alist nth)
      (mew-assoc-equal key alist nth))))

(defun mew-assoc-match (key alist nth)
  (if (null alist) 
      ()
    (let* ((car (car alist))
	   (nthobj (nth nth car)))
      (if (or (and (stringp nthobj)
		   (let ((case-fold-search t))
		     (string-match key nthobj))) ;;; here here
	      (equal nthobj key)
	      (eq nthobj t))
	  car
	(mew-assoc-match key (cdr alist) nth)))))

(defun mew-assoc2 (key alist nth match)
  (let ((case-fold-search t))
    (if match
	(mew-assoc-match2 key alist nth)
      (mew-assoc-equal key alist nth))))

(defun mew-assoc-match2 (key alist nth)
  (if (null alist) 
      ()
    (let* ((car (car alist))
	   (nthobj (nth nth car)))
      (if (or (and (stringp nthobj)
		   (let ((case-fold-search t))
		     (string-match nthobj key))) ;;; here here
	      (equal nthobj key)
	      (eq nthobj t))
	  car
	(mew-assoc-match2 key (cdr alist) nth)))))

(defun mew-assoc-equal (key alist nth)
  (if (null alist) 
      ()
    (let* ((car (car alist))
	   (nthobj (nth nth car)))
      (if (or (equal nthobj key)
	      (eq nthobj t))
	  car
	(mew-assoc-equal key (cdr alist) nth)))))

;;(defun mew-assoc2 (key alist nth exact)
;;  (let ((case-fold-search t))
;;    (cond
;;     ((null alist) ())
;;     ((and exact 
;;	   (stringp (nth nth (car alist)))
;;	   (equal (downcase key) (downcase (nth nth (car alist)))))
;;      (car alist))
;;     ((equal (nth nth (car alist)) key)
;;      (car alist))
;;     ((eq (nth nth (car alist)) t)
;;      (car alist))
;;     (t (mew-assoc key (cdr alist) nth exact))
;;     )
;;    ))

(defun mew-delete-line ()
  (beginning-of-line)
  (let ((start (point)))
    (forward-line)
    (delete-region start (point))
    ))

(defmacro mew-directory-empty-p (dir)
  (` (null (car (cdr (cdr (directory-files (, dir)))))))
  )

(defun mew-which(file path)
  (catch 'loop
    (while path
      (if (file-exists-p (expand-file-name file (car path)))
	  (throw 'loop (expand-file-name file (car path)))
	(setq path (cdr path)))
      )
    ))

;;;
;;; Window configuration stack
;;;

(defvar mew-window-stack nil)

(defun mew-window-push ()
  (interactive)
  (let* ((key (cond
	       ((and mew-emacs19-p (not mew-xemacs-p))
		(selected-frame))
	       (t 'dummy)
	       ))
	 (stack (mew-alist-search mew-window-stack key)))
    (setq stack (cons (current-window-configuration) stack))
    (setq mew-window-stack
	  (mew-alist-add mew-window-stack key stack))
    ))

(defun mew-window-pop ()
  (interactive)
  (let* ((key (cond
	       ((and mew-emacs19-p (not mew-xemacs-p))
		(selected-frame))
	       (t 'dummy)
	       ))
	 (stack (mew-alist-search mew-window-stack key)))
    (if stack
	(set-window-configuration (car-safe stack))
      (if (and mew-emacs19-p
	       (not mew-xemacs-p)
	       (nth 1 (frame-list))
	       (mew-y-or-n-p "Delete this frame? "))
	  (delete-frame)
	(message "Window Stack is empty.")))
    (setq stack (cdr-safe stack))
    (setq mew-window-stack
	  (mew-alist-add mew-window-stack key stack))
    ))

;;
;; folders
;;

(defmacro mew-folder-make-alist (list)
  (` (mapcar (function mew-folder-pair) (, list))))

;; never defmacro for (mapcar)
(defun mew-folder-pair (string)
  (cons (format "+%s" string) (format "+%s" string)))

(defmacro mew-refile-make-alist (list)
  (` (mapcar (function mew-refile-pair) (, list))))
(defun mew-refile-pair (string)
  (if (mew-folders-ignore-p string)
      nil
    (cons (downcase (file-name-nondirectory (directory-file-name string)))
	  (format "+%s" (directory-file-name string)))

    ))

(defun mew-folders-ignore-p (string)
  (let ((ilist mew-folders-ignore))
    (catch 'ignore
      ;; while always returns nil
      (while ilist
	(if (string-match (concat "^" (car ilist))
			  string)
	    (throw 'ignore t))
	(setq ilist (cdr ilist))))))

(defun mew-folder-make-list ()
  (save-excursion
    (let ((case-fold-search t)
	  (folders ())
	  (folder nil)
	  (start nil))
      (set-buffer (get-buffer-create mew-buffer-tmp))
      (erase-buffer)
      (if (file-exists-p mew-folders-file)
	  (insert-file-contents mew-folders-file)
	(call-process mew-prog-folders nil t nil "-fast" "-recurse")
	)
      (goto-char (point-min))
      (while (not (eobp))
        (setq start (point))
        (forward-line 1)
	(setq folder (mew-buffer-substring start (1- (point))))
	(if (string-match (concat "^" (car folders) "/") folder)
	    (setq folders 
		  (cons folder 
			(cons (concat (car folders) "/"	)
			      (cdr folders))))
	  (setq folders (cons folder folders)))
	)
      folders ;; return value
      )))

;;;
;;; Bootstrap --- mew and mew-send
;;;

(defun mew (&optional arg)
  (interactive "P")
  (mew-window-push)
  (if (null mew-path) (mew-init))
  (if arg
      (mew-summary-goto-folder) ;; C-u
    (let ((folder mew-inbox))
      (if (get-buffer folder)
	  (switch-to-buffer folder)
	(mew-summary-folder-create folder))
      (mew-summary-inc)
      )
    ))

(defun mew-send (&optional to cc subject)
  (interactive)
  (mew-current-set 'window (current-window-configuration))
  (if (null mew-path) (mew-init))
  (mew-summary-send t to cc subject)
  )

;;
;; Functions for boot time
;;

(defun mew-init ()
  (mew-hello)
  (if mew-demo
      (mew-demo))
  (message "Setting mew world ...")
  (mew-set-environment)
  (if mew-use-bbdb (require 'bbdb-com)) ;; bbdb is implicitly required
  (run-hooks 'mew-init-hook)
  (message "Setting mew world ... done")
  )

(defun mew-center-line ()
  (save-excursion
    (end-of-line)
    (let ((ll (current-column)))
      (beginning-of-line)
      (indent-to (/ (- (window-width) ll) 2))
      )
    ))

(defun mew-center-region (from to)
  (interactive "r")
  (if (> from to)
      (let ((tem to))
        (setq to from from tem)))
  (save-excursion
    (save-restriction
      (narrow-to-region from to)
      (goto-char from)
      (while (not (eobp))
        (mew-center-line)
        (forward-line 1)))))

(defun mew-hello ()
  (let ((nbuf (get-buffer-create mew-buffer-hello)))
    (mew-window-configure nbuf '(1 0))
    (erase-buffer)
    (insert (format mew-hello-message mew-version))
    (mew-center-region (point-min) (point-max))
    (end-of-line)
    (insert (make-string (1- (- (window-width) (current-column))) 32))
    (sit-for 0) ;; to redraw
    ))

(defun mew-set-environment ()
  (let ((xerr nil))
    (condition-case err
	(progn
	  (set-buffer (get-buffer-create mew-buffer-tmp))
	  (erase-buffer)
	  (if (file-exists-p mew-profile)
	      (insert-file-contents mew-profile)
	    (setq xerr (format "No %s" mew-profile))
	    (error ""))
	  (if (setq mew-path (mew-header-get-value "Path:"))
	      (setq mew-path
		    (expand-file-name 
		     (mew-header-delete-ws mew-path)
		     (expand-file-name "~")))
	    (setq xerr "No Path:")
	    (error ""))
	  (if (setq mew-draft-folder (mew-header-get-value "Draft-Folder:"))
	      (setq mew-draft-folder
		    (format "+%s" (mew-folder-to-dir
				   (mew-header-delete-ws		  
				    mew-draft-folder))))
	    (setq xerr "No Draft-Folder:")
	    (error ""))
	  ;; mew-temp-file must be local and readable for the user only
	  ;; for privacy/speed reasons.
	  (setq mew-temp-dir (make-temp-name mew-temp-file-initial))
	  (mew-make-directory mew-temp-dir)
	  (call-process mew-prog-chmod nil nil nil 
			mew-prog-chmod-arg mew-temp-dir)
	  (setq mew-temp-file (expand-file-name "mew" mew-temp-dir))
	  ;;
	  (set-buffer (get-buffer-create mew-buffer-tmp))
	  (erase-buffer)
	  (call-process mew-prog-mime-encode nil t t "-v")
	  (goto-char (point-min))
	  (if (re-search-forward "version 0.08 961012" nil t)
	      ()
	    (setq xerr "mewencode version mismatch")
	    (error ""))
	  ;;
	  (erase-buffer)
	  (call-process mew-prog-msgchk nil t t "-help")
	  (goto-char (point-min))
	  (if (re-search-forward "6\\.8\\.3\\.JP" nil t)
	      ;; xxx this will be soon obsolete...
	      (setq mew-header-decode-regex 
		    "=\\?\\([^?]+\\)\\?\\(.\\)\\?\\([^?]+\\)\\?= ?")
	    (setq mew-header-decode-regex 
		  "=\\?\\([^?]+\\)\\?\\(.\\)\\?\\([^?]+\\)\\?="))
	  ;;
	  (if (null mew-alias-alist)
	      (setq mew-alias-alist (mew-alias-make-alist)))
	  ;;
	  (setq mew-folders-file 
		(expand-file-name mew-folders-file mew-path))
	  (cond 
	   ((null mew-folder-list)
	    (setq mew-folder-list (mew-folder-make-list))
	    (setq mew-folder-alist (mew-folder-make-alist mew-folder-list))
	    (setq mew-refile-alist (mew-refile-make-alist mew-folder-list))
	    )
	   )
	  )
      (error (set-buffer mew-buffer-hello)
	     (insert mew-error-message)
	     (set-buffer-modified-p nil)
	     (setq buffer-read-only t)
	     (if xerr (error xerr)
	       (error "If you have ~/Mail/.mew-from-alist, remove it. Or maybe exec-path error")))
      )
    ))

;;;
;;; window configuration
;;;

(defun mew-window-configure (nbuf action)
;;; action : summary, message or list
;;; list for action (1 0)  for Summary only
;;; list for action (3 10) for Summary and Message
  (if (equal action 'summary)
      (mew-current-set 'message nil))
  (let* ((windows
	  (if (listp action) 
	      action
	    (car (cdr (assq action mew-window-configuration)))))
	 (obufwin (get-buffer-window (current-buffer)))
	 (msgwin  (get-buffer-window (mew-buffer-message)))
	 (height nil) (winsum nil) (sumheight 0) (msgheight 0))
    (cond 
     (mew-window-use-full
      ;; Delete other windows and use full emacs window.
      (delete-other-windows)
      (setq height (window-height (selected-window))))
     (t
      (setq height
	    (+ (if obufwin (window-height obufwin) 0)
	       (if msgwin  (window-height msgwin)  0)
	       ))))
    (if (get-buffer (mew-buffer-message))
	(delete-windows-on (mew-buffer-message))
      (save-excursion
	(set-buffer (get-buffer-create (mew-buffer-message)))
	(mew-message-mode)))
    (setq winsum (apply (function +) windows))
    (if (not (zerop (nth 0 windows)))
	(setq sumheight (max window-min-height
			     (/ (* height (nth 0 windows)) winsum))))
    (if (equal (% sumheight 2) 1) (setq sumheight (1+ sumheight)))
    (if (not (zerop (nth 1 windows)))
	(setq msgheight (max window-min-height
			     (- height sumheight))))
    (setq height (+ sumheight msgheight))

    (if (or (not mew-xemacs-p)
	    ;; xemacs doesn't like to enlarge the only window:
	    (> (length (window-list)) 2))
	(enlarge-window (max 0 (- height (window-height (selected-window))))))

    (if (null (zerop sumheight))
	(switch-to-buffer nbuf 'norecord))
    (if (zerop msgheight)
	()
      (split-window nil sumheight)
      (other-window 1)
      (switch-to-buffer (mew-buffer-message) 'norecord))
    ))

;;;
;;; Input method
;;;

(defun mew-input-folder (default)
  (let ((folder))
    (setq folder (completing-read (format "Folder name (%s): " default)
				  mew-folder-alist
				  nil
				  nil  ;; not require match
				  "+"))
    (directory-file-name 
     (if (or (string= folder "") (string= folder "+"))
 	 default
       folder))))

(defun mew-input-folders (default &optional string)
  (let ((folders))
    (setq folders
	  (read-from-minibuffer (format "Folder name (%s): " default)
				(or string "+")
				mew-folder-map))
    (if (or (string= folders "") (string= folders "+"))
	(setq folders default))
    (mapcar (function directory-file-name) 
	    (mew-header-split
	     (mew-header-delete-ws folders) ?,))
    ))

(defun mew-input-range (folder)
  "Return (range erase-update)."
  (mew-summary-multipart-delete)
  ;; for the case when parts are expanded in the bottom of the folder
  (let*
      ((pair (assoc folder mew-prog-scan-default-range-alist))
       (default (or (cdr pair) mew-prog-scan-default-range))
       (range (if mew-ask-range
		  (read-string (format "Range (%s): " default) "")
		"")))
    (if (string= range "")
	(setq range default))
    (cond
     ;; range is other than "update"
     ((not (string= range "update"))
      (list range 'erase)) ;; non-update, erase it
     ;; update
     ((get-buffer folder)
      (save-excursion
	(set-buffer folder)
	(goto-char (point-max))
	(if (bobp)
	    (list "all" 'update) ;; buffer is empty. no need to erase
	  (forward-line -1)
	  (list 
	   (concat
	    (int-to-string (1+ (string-to-int (mew-summary-message-number))))
	    "-" 
	    "last")
	   'update) ;; this is update!
	  )))
     ;; update but folder doesn't exist in Emacs. 
     (t (list "all" 'update)) ;; no need to erase
     )
    ))

(defun mew-input-address (prompt)
  (read-from-minibuffer (concat prompt " ") "" mew-minibuffer-map nil)
  )

(defun mew-input-address2 (prompt default)
  (let ((ret (read-from-minibuffer 
	      (format prompt default) "" mew-minibuffer-map nil)))
    (if (string= ret "") default ret)
    ))

(defun mew-input-filename (&optional prompt default)
  (let ((msg (or prompt "File: "))
	(file (concat "~/" default)))
    (expand-file-name (read-file-name msg file))
    ))

(defun mew-input-directory-name ()
  (let ((dir (expand-file-name (read-file-name "Directory : " nil "~" t))))
    (if (file-directory-p dir)
	dir
      (mew-input-directory-name)
      )
    ))

(defun mew-input-string (prompt subdir default)
  (let ((input (read-string (format prompt subdir default) "")))
    (if (string= input "") default input))
  )

(defun mew-input-type (prompt filename default type-list)
  (let ((completion-ignore-case t)
	(type))
    (setq type (completing-read
		(format prompt filename default)
		(mapcar (function (lambda (x) (cons x x)))
			type-list)
		nil
		t  ;; not require match
		""))
    (if (string= type "") default type)
    ))

;;;
;;; Misc
;;;

(defun mew-boundary-get (&optional string)
  (if (null string) (setq string "Next_Part"))
  (format mew-default-boundary
	  string
	  (mew-replace-character (current-time-string) 32 ?_) ; 32 == " "
	  )
  )

(defun mew-expand-file-name (msg &optional buffer-or-name)
  (if (bufferp buffer-or-name)
      (setq buffer-or-name (buffer-name buffer-or-name)))
  (and buffer-or-name 
       (setq buffer-or-name (directory-file-name buffer-or-name)))
  (if buffer-or-name
      (expand-file-name 
       (concat (mew-folder-to-dir buffer-or-name) "/" (mew-folder-to-dir msg))
       mew-path)
    (expand-file-name (mew-folder-to-dir msg) mew-path))
  )

;;
;; e-mail address alias
;;

(defun mew-alias-expand (alias)
  "The function of MH alias expansion. If ALIAS can be expanded, it 
returns the expansion. Otherwise, it returns ALIAS itself."
  (let ((expn (cdr (assoc alias mew-alias-alist))))
    ;assoc returns nil if alist is nil
    (if expn expn alias)))

(defun mew-alias-make-alist ()
  "Make alias alist with association of (alias . expantion) from
MH \"ali\" command. Currently, only \"user: user@domain\" syntax
is supported."
  (save-excursion
    (let ((case-fold-search t)
	  (alias nil)
	  (expn nil)
	  (alist nil)
	  (buf (get-buffer-create mew-buffer-tmp)))
      (set-buffer buf)
      (erase-buffer)
      (call-process mew-prog-ali nil t nil "-nolist" "-nonormalize" "-nouser")
      ;; concat separated lines by comma
      (goto-char (point-min))
      (while (re-search-forward ",$" nil t)
	(end-of-line)
	(forward-char 1)
 	(delete-backward-char 1))
      ;;
      (goto-char (point-min))
      (while (re-search-forward "^\\([^:]+\\):\\s-+\\(.*\\)$" nil t)
	(setq alias (mew-match 1)
	      expn (mew-match 2))
	;; append for first assoc comes first
	(setq alist (cons (cons alias expn) alist)))
      ;; load from-alist
      (if (not mew-from-alist)
	  (setq mew-from-alist
		(mew-alist-load mew-from-file-name)))
      (mapcar
       '(lambda (arg)
	  (if (and (car arg)
		   (not (assoc (mew-header-delete-at (car arg)) alist)))
	      (setq alist 
		    (cons (cons (mew-header-delete-at (car arg)) 
				(car arg))
			  alist))))
       mew-from-alist)
      (nreverse alist) ; return value
      )))

;;
;;
;;

(defun mew-mime-start-process (begin end program options)
  (if (mew-which program exec-path)
      (if (mew-y-or-n-p (format "Start %s? " program))
	  (let ((mc-flag nil)
		(file-coding-system (if mew-mule-p *noconv*))
		pro file)
	    (if (not (file-exists-p mew-temp-dir))
		(mew-make-directory mew-temp-dir)) ;; just in case
	    (setq file (make-temp-name mew-temp-file)) 
	    (write-region begin end file)
	    (message "Starting %s ..." program)
	    (setq pro (apply (function start-process)
			     (format "*mew %s*" program)
			     mew-buffer-tmp ;; xxx
			     program
			     (append options (list file))))
	    (set-process-sentinel pro 'mew-mime-start-process-sentinel)
	    (message "Sending %s ... done" program)
	    (setq mew-process-file-alist
		  (cons (cons pro file) mew-process-file-alist))
	    ))
    (message "Program %s is not found" program))
  t ;; to next part
  )

(defun mew-mime-start-process-sentinel (process event)
  (let ((al (assoc process mew-process-file-alist)))
    (if (cdr al) (delete-file (cdr al)))
    (setq mew-process-file-alist (mew-delq al mew-process-file-alist))
    ))

(defun mew-mime-call-process (begin end program options)
  (if (mew-which program exec-path)
      (if (mew-y-or-n-p (format "Call %s? " program))
	  (let ((mc-flag nil)
		(file-coding-system (if mew-mule-p *noconv*))
		file)
	    (message "Calling %s ..." program)
	    (if (not (file-exists-p mew-temp-dir))
		(mew-make-directory mew-temp-dir)) ;; just in case
	    (setq file (make-temp-name mew-temp-file))
	    ;; NEVER use call-process-region for privary reasons
	    (write-region begin end file)
	    (apply (function call-process) program file nil nil options)
	    (message "Calling %s ... done" program)
	    ))
    (message "Program %s is not found" program))
  t ;; to next part
  )

(defun mew-mime-prev-page ()
  (interactive)
  (let ((buf (current-buffer)))
    (unwind-protect
	(progn
	  (pop-to-buffer (mew-buffer-message))
	  (scroll-down))
      (pop-to-buffer buf)))
  )

(defun mew-message-header-end (begin end)
  (save-excursion
    (set-buffer (mew-current-get 'cache))
    (narrow-to-region begin end)  
    (goto-char (point-min))
    (re-search-forward "^-*$")
    (save-restriction ;; for no body
      (forward-line 1))
    (prog1 
	(point)
      (widen))
    ))

;; change
(defun mew-mime-text/plain (begin end &optional params)
  (if (> end begin)
      (save-excursion
	(set-buffer (mew-buffer-message))
	(let ((buffer-read-only nil))
	  (insert-buffer-substring (mew-current-get 'cache) begin end)
	  (if mew-eof-string
	      (progn
		(goto-char (point-max)) ;; necessary?
		(insert mew-eof-string)))
	  )
	(cond (mew-break-pages 
	       (goto-char (point-min))
	       (mew-message-narrow-to-page)
	       ))
	(set-buffer-modified-p nil) ;; xxx
	))
  )

(defun mew-mime-application/octet-stream (begin end &optional params)
  (if (not mew-save-application/octet-stream-p)
      (mew-mime-text/plain begin end params)
    (if (not (mew-y-or-n-p "Save this part? "))
	()
      (save-excursion
	(set-buffer (mew-current-get 'cache))
	(let* ((name (mew-syntax-get-member params "name"))
	       (file (mew-input-filename nil name))
	       (mc-flag nil)
	       (file-coding-system (if mew-mule-p *noconv*)))
	  (if (file-exists-p file)
	      (if mew-file-append-p
		  (write-region begin end file t)
		(if (mew-y-or-n-p (format "File %s exists. Append it? "
					  (file-name-nondirectory file)))
		    (write-region begin end file t)
		  ()))
	    (write-region begin end file)
	    ))
	))
    ))

;;
;;
;;

(defmacro mew-cache-mime-syntax (buf)
  (` (save-excursion
       (set-buffer (, buf))
       mew-decode-syntax)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Prepare new message --- caching
;;

(defvar mew-cache nil
      "Mail cache. (old ... new) order alist with association
 ((\"+folder\" . \"message\") . cache-buffer)"
      )

(defmacro mew-cache-buffer-get (entry)
  (` (cdr (, entry))))

(defmacro mew-cache-folder-get (entry)
  (` (car (car (, entry)))))

(defmacro mew-cache-message-get (entry)
  (` (cdr (car (, entry)))))

(defmacro mew-cache-entry-make (fld-msg buf)
  (` (cons (, fld-msg) (, buf))))

(defmacro mew-cache-hit (fld-msg)
  "Return value assosiated with key."
  (` (mew-cache-buffer-get (assoc (, fld-msg) mew-cache))))

(defun mew-cache-sort (entry)
  (let* ((pointer (cons nil mew-cache))
	 (top pointer))
    (while (cdr pointer)
      (if (equal (car (cdr pointer)) entry)
	  (setcdr pointer (cdr (cdr pointer)))
	(setq pointer (cdr pointer))))
    (setcdr pointer (list entry))
    (setq mew-cache (cdr top))))

(defun mew-cache-add (fld-msg)
  "Rutern associating buffer"
  (let ((len (length mew-cache))
	(buf nil))
    (if (>= len mew-cache-size)
	(prog1
	    (setq buf (mew-cache-buffer-get (car mew-cache)))
	  (setq mew-cache (nconc (cdr mew-cache)
				  (list (mew-cache-entry-make fld-msg buf))))
	  )
      (prog1
	  (setq buf (get-buffer-create 
		     (concat mew-buffer-cache (int-to-string len))))
	(setq mew-cache (nconc mew-cache
				(list (mew-cache-entry-make fld-msg buf))))
	))
    ))

(defun mew-cache-attribute-get (file)
  ;; 5th is modified time
  ;; 7th is file size
  (let ((attr (file-attributes file)))
    (list (nth 5 attr)
	  (nth 7 attr))))
       
(defun mew-cache-message (fld-msg)
  (let ((hit (mew-cache-hit fld-msg))
	(file (mew-expand-file-name (concat (car fld-msg) "/" (cdr fld-msg)))))
    (save-excursion
      (if hit
	  (progn
	    (mew-cache-sort (mew-cache-entry-make fld-msg hit))
	    (set-buffer hit)
	    ;; in cache buffer
	    (if (null (equal mew-cache-attribute 
			     (mew-cache-attribute-get file)))
		;; cache is invalid
		(mew-cache-message-new file)))
	;; no cache hit
	(setq hit (mew-cache-add fld-msg))
	(set-buffer hit)
	;; in cache buffer
	(mew-cache-message-new file)
	)
      hit ;; retrun value
      )))

;; use to debug
(defun mew-cache-flush ()
  (interactive)
  (mew-summary-multipart-delete)
  (let ((n 0))
    (while (< n mew-cache-size)
      (mew-kill-buffer (concat mew-buffer-cache (int-to-string n)))
      (setq n (1+ n))
      ))
  (mew-current-set 'cache nil)
  (setq mew-cache nil)
  )

;;
;; Cache header coordination
;;

(defun mew-cache-delete-field (field begin)
  (let ((case-fold-search t)
	(start nil))
    (save-excursion
      (goto-char (or begin (point-min)))
      (if (re-search-forward (format "^$\\|^%s" field) nil t)
	  (if (not (string= (downcase (mew-match 0)) (downcase field)))
	      nil
	    (setq start (match-beginning 0))
	    (forward-line 1)
	    (while (looking-at "[ \t]")
	      (forward-line 1))
	    (prog1
		(mew-buffer-substring start (point))
	      (delete-region start (point)))
	    )
	nil)
      )))

(defun mew-cache-delete-content-fields (&optional begin mime)
  (let ((case-fold-search t)
	(start nil)
	key)
    (save-excursion
      (goto-char (or begin (point-min)))
      (catch 'loop
	(while (re-search-forward "^$\\|^Content-" nil t)
	  (setq key (mew-match 0))
	  (setq start (match-beginning 0))
	  (if (string-match "^$" key) 
	      (if (not mime)
		  (throw 'loop nil)
		(forward-line 1)
		(delete-region start (point))
		(throw 'loop nil)
		))
	  (forward-line 1)
	  (while (looking-at "[ \t]")
	    (forward-line 1))
	  (delete-region start (point))
	  ))
      )))

;;;
;;;
;;;

(defun mew-header-insert-from-file (file)
  (insert-file-contents file) ;; whole message
  ;;(insert "\n")
  (mew-header-arrange)
  ;; Some nice features for XEmacs.
  (if mew-xemacs-p (highlight-headers (point-min) (point-max) nil))
  )

(defun mew-header-insert-from-cache (cache begin end)
  (insert-buffer-substring cache begin end)
  (mew-header-arrange)
  ;; Some nice features for XEmacs.
  (if mew-xemacs-p (highlight-headers (point-min) (point-max) nil))
  (if mew-decode-result (insert mew-decode-result))
  (insert "\n")
  )

;;;
;;; mini buffer
;;;

(defun mew-minibuffer-alias ()
  (interactive)
  (let ((word nil) (comp nil) (all nil) (len nil))
    (if (null (setq word (mew-delete-backward-char)))
	(tab-to-tab-stop) ;; default key binding
      (setq comp (try-completion word mew-alias-alist))
      (setq all (all-completions word mew-alias-alist))
      (setq len (length word))
      (cond
       ((eq comp t)
	(insert (cdr (assoc word mew-alias-alist)))
	(and (get-buffer mew-buffer-completions)
	     (kill-buffer mew-buffer-completions)
	     (mew-window-pop)))
       ((and (equal (aref word (1- len)) ?@)
	     (assoc (substring word 0 (1- len)) mew-alias-alist))
	(insert (cdr (assoc (substring word 0 (1- len)) mew-alias-alist)))
	(and (get-buffer mew-buffer-completions)
	     (kill-buffer mew-buffer-completions)
	     (mew-window-pop)))
       ((stringp comp)
	(insert comp)
	(and (> (length all) 1)
	     (mew-window-push)
	     (with-output-to-temp-buffer mew-buffer-completions
	       (display-completion-list all))))
       (t (insert word)
	  (mew-temp-minibuffer-message "No matching alias"))
       )
      )
    ))

(defun mew-minibuffer-domain ()
  (interactive)
  (let ((word (mew-delete-backward-char-at)))
;;;;    (if (null (setq word (mew-delete-backward-char-at)))
;;;;	(lisp-complete-symbol) ;; default key binding)
    ;; word is strings between @ and cursor
    (cond
     ((equal word nil) ;; @ doesn't exist.
      (if (null mew-mail-domain-list)
	  ()
	(insert "@")
	(with-output-to-temp-buffer mew-buffer-completions
	  (display-completion-list mew-mail-domain-list))))
     ((equal word t) ;; just after @
      (if mew-mail-domain-list
	  (with-output-to-temp-buffer mew-buffer-completions
	    (display-completion-list mew-mail-domain-list)))
      )
     (t
      (let ((comp nil) (all nil) (len nil)
	    (dalist (mew-slide-pair mew-mail-domain-list)))
	(setq comp (try-completion word dalist))
	(setq all (all-completions word dalist))
	(setq len (length word))
	(cond
	 ;; already completed
	 ((eq comp t) 
	  (insert (cdr (assoc word dalist))) ;; next candidate
	  (and (get-buffer mew-buffer-completions)
	       (kill-buffer mew-buffer-completions))
	  )
	 ;; just one candidate
	 ((equal 1 (length all))
	  (insert (car all))
	  (and (get-buffer mew-buffer-completions)
	       (kill-buffer mew-buffer-completions))
	  (mew-temp-minibuffer-message "Sole completion")
	  )
	 ;; one or more candidate
	 ((stringp comp)
	  (insert comp)
	  (and (> (length all) 1)
	       (with-output-to-temp-buffer mew-buffer-completions
		 (display-completion-list all))))
	 ;; no candidate
	 (t (insert word)
	    (mew-temp-minibuffer-message "No matching domain"))
	 ))
      )
     )
    ))

;;
;; folder completion
;;

(defun mew-temp-minibuffer-message (m)
  (let ((savemax (point-max)))
    (save-excursion
      (goto-char (point-max))
      (insert m))
    (let ((inhibit-quit t))
      (sit-for 2)
      (delete-region savemax (point-max))
      (if quit-flag (setq quit-flag nil	unread-command-char 7))
      )))

(defun mew-minibuffer-folders ()
  (interactive)
  (let ((word nil) (comp nil) (all nil) (len nil))
    (if (null (setq word (mew-delete-backward-char)))
	(tab-to-tab-stop) ;; default key binding
      (setq comp (try-completion word mew-folder-alist))
      (setq all (all-completions word mew-folder-alist))
      (setq len (length word))
      (cond
       ((eq comp t) 
	;; insert completed word
	(insert (cdr (assoc word mew-folder-alist)))
	(and (get-buffer mew-buffer-completions)
	     (kill-buffer mew-buffer-completions)))
       ((stringp comp)
	(insert comp)
	(if (string-equal comp word)
	    (mew-temp-minibuffer-message " [Complete, but not unique]")))
       (t (insert word)
	  (mew-temp-minibuffer-message " No matching folder"))
       )
      )
    ))

(defun mew-minibuffer-folders-word ()
  (interactive)
  (let ((word nil) (comp nil) (all nil) (len nil))
    (if (null (setq word (mew-delete-backward-char)))
	(tab-to-tab-stop) ;; default key binding
      (setq comp (try-completion word mew-folder-alist))
      (setq all (all-completions word mew-folder-alist))
      (setq len (length word))
      (cond
       ((eq comp t) 
	;; insert completed word
	(insert (cdr (assoc word mew-folder-alist)))
	(and (get-buffer mew-buffer-completions)
	     (kill-buffer mew-buffer-completions)))
       ((stringp comp)
	(insert comp) ;; insert substring
	(and (> (length all) 1)
	     (with-output-to-temp-buffer mew-buffer-completions
	       (display-completion-list all))))
       (t (insert word)
	  (mew-temp-minibuffer-message " No matching folder"))
       )
      )
    ))

;;

(defun mew-clean-up ()
  (if (and mew-temp-dir (file-exists-p mew-temp-dir))
      (call-process "rm" nil nil nil "-rf" mew-temp-dir))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;
(require 'mew-charset)
(require 'mew-header)
(require 'mew-pgp)
(require 'mew-encode)
(require 'mew-decode)
(require 'mew-syntax)
(require 'mew-scan)
(require 'mew-pick)
(require 'mew-summary)
(require 'mew-virtual)
(require 'mew-message)
(require 'mew-draft)
(require 'mew-attach)
(require 'mew-demo)
(require 'mew-refile)
(require 'mew-ext)
(require 'mew-fib)

